#include <stdio.h>
#include <stdarg.h>
#include <ntddk.h>

#define NR_DEVICES	8

struct irp_list {
	IRP *irp;
	LIST_ENTRY lh;
};

struct dev_info {
	LIST_ENTRY lh;
	FAST_MUTEX mutex;
};

NTSTATUS open(PDEVICE_OBJECT device, IRP *irp)
{
	irp->IoStatus.Status=STATUS_SUCCESS;
	irp->IoStatus.Information=0;
	IoCompleteRequest(irp, IO_NO_INCREMENT);
	return STATUS_SUCCESS;
}


NTSTATUS close(PDEVICE_OBJECT device, IRP *irp)
{
	irp->IoStatus.Status=STATUS_SUCCESS;
	irp->IoStatus.Information=0;
	IoCompleteRequest(irp, IO_NO_INCREMENT);
	return STATUS_SUCCESS;
}



NTSTATUS cleanup(PDEVICE_OBJECT device, IRP *irp)
{
	LIST_ENTRY *i, *next;
	struct dev_info *di=(struct dev_info *)device->DeviceExtension;

	ExAcquireFastMutex(&di->mutex);
	/* fetch i->Next in advance to avoid trouble when freeing i */
	for(i=di->lh.Flink; i != &di->lh && (next=i->Flink,1); i=next) {
		struct irp_list *il=CONTAINING_RECORD(i, struct irp_list, lh);
		IO_STACK_LOCATION *isl;
		
		isl=IoGetCurrentIrpStackLocation(il->irp);
		
		if (isl->FileObject == IoGetCurrentIrpStackLocation(irp)->FileObject) {
			il->irp->IoStatus.Status=STATUS_CANCELLED;
			IoCompleteRequest(il->irp, IO_NO_INCREMENT);
			RemoveEntryList(i);
			/* here we might loose i->Next */
			ExFreePool(il);
		}
	}
	ExReleaseFastMutex(&di->mutex);

	irp->IoStatus.Status=STATUS_SUCCESS;
	irp->IoStatus.Information=0;
	IoCompleteRequest(irp, IO_NO_INCREMENT);
	return STATUS_SUCCESS;
}


NTSTATUS write(PDEVICE_OBJECT device, IRP *irp)
{
	struct dev_info *di=(struct dev_info*)device->DeviceExtension;
	struct irp_list *il;
	NTSTATUS ret;

	if (!(il=ExAllocatePool(PagedPool, sizeof(*il)))) {
		ret=STATUS_NO_MEMORY;
		goto error;
	}
	il->irp=irp;
	
	ExAcquireFastMutex(&di->mutex);
	InsertTailList(&di->lh, &il->lh);
	irp->IoStatus.Information=0;
	IoMarkIrpPending(irp);
	ExReleaseFastMutex(&di->mutex);
	

	return STATUS_PENDING;

error:
	irp->IoStatus.Status=ret;
	irp->IoStatus.Information=0;
	IoCompleteRequest(irp, IO_NO_INCREMENT);
	return ret;
}

NTSTATUS read(PDEVICE_OBJECT device, IRP *irp)
{
	struct dev_info *di=(struct dev_info*)device->DeviceExtension;
	LIST_ENTRY *i, *next;
	char *buffer;
	int size, k, l=0;

	buffer=irp->AssociatedIrp.SystemBuffer;
	size=IoGetCurrentIrpStackLocation(irp)->Parameters.Read.Length;

	ExAcquireFastMutex(&di->mutex);

	/* fetch the i->Next in advance to avoid trouble when freeing i*/
	for(i=di->lh.Flink; i != &di->lh && (next=i->Flink,1); i=next) {
		struct irp_list *il=CONTAINING_RECORD(i, struct irp_list, lh);
		const char *il_buffer;
		int il_size;
		
		il_size=IoGetCurrentIrpStackLocation(il->irp)->Parameters.Write.Length;
		il_buffer=il->irp->AssociatedIrp.SystemBuffer;

		for(k=il->irp->IoStatus.Information; k < il_size && l < size; k++, l++) {
			buffer[l]=(il_buffer[k] >= 'A' && il_buffer[k] <= 'Z')?
				il_buffer[k]-'A'+'a':il_buffer[k];
		}

		il->irp->IoStatus.Information=k;
		
		if (k == il_size) {
			RemoveEntryList(i);
			il->irp->IoStatus.Status=STATUS_SUCCESS;
			IoCompleteRequest(il->irp, IO_NO_INCREMENT);
			/* here we might lose i->Next */
			ExFreePool(il);
		} else


		if (l == size) 
			break;
	}
	ExReleaseFastMutex(&di->mutex);

	irp->IoStatus.Status=STATUS_SUCCESS;
	irp->IoStatus.Information=l;
	IoCompleteRequest(irp, IO_NO_INCREMENT);
	return STATUS_SUCCESS;
}

UNICODE_STRING* TO_UNICODE(const char *str, UNICODE_STRING *ustr) 
{
	ANSI_STRING astr;

	RtlInitAnsiString(&astr, str);
	if (RtlAnsiStringToUnicodeString(ustr, &astr, TRUE) != STATUS_SUCCESS) 
		return NULL;

	return ustr;
}



void DriverUnload(PDRIVER_OBJECT driver)
{
	DEVICE_OBJECT *device;
	int i;

	for(i=0; i<NR_DEVICES; i++) {
		char link_name[]={'\\','?','?','\\','c','a','s','e','0'+i,0};
		UNICODE_STRING link_uname;

		if (TO_UNICODE(link_name, &link_uname)) 
			IoDeleteSymbolicLink(&link_uname);

		if (link_uname.Buffer)
			RtlFreeUnicodeString(&link_uname);
	}


	while ((device=driver->DeviceObject))
		IoDeleteDevice(device);
}



NTSTATUS DriverEntry(PDRIVER_OBJECT  driver, PUNICODE_STRING  registry)
{
	NTSTATUS ret;
	struct dev_info *di;
	DEVICE_OBJECT *device;
	int i;
	UNICODE_STRING dev_uname, link_uname;


	memset(&dev_uname, 0, sizeof(dev_uname)); 
	memset(&link_uname, 0, sizeof(link_uname));

	
	for(i=0; i<NR_DEVICES; i++) {
		char dev_name[]={'\\','D','e','v','i','c','e','\\','c','a','s','e','0'+i,0};
		char link_name[]={'\\','?','?','\\','c','a','s','e','0'+i,0};
		

		if ((ret=IoCreateDevice(driver, sizeof(*di), TO_UNICODE(dev_name, &dev_uname), 
			0xF0000000,0, FALSE, &device)) != STATUS_SUCCESS)
			goto error;

		if  ((ret=IoCreateSymbolicLink(TO_UNICODE(link_name, &link_uname), &dev_uname)) != STATUS_SUCCESS)
			goto error;
		
		
		device->Flags|=DO_BUFFERED_IO;
		di=(struct dev_info*)device->DeviceExtension;
		InitializeListHead(&di->lh);
		ExInitializeFastMutex(&di->mutex);

		if (dev_uname.Buffer)
			RtlFreeUnicodeString(&dev_uname);
		if (link_uname.Buffer) 
			RtlFreeUnicodeString(&link_uname);

	}

	driver->DriverUnload=DriverUnload;
	driver->MajorFunction[IRP_MJ_CREATE]=open;
	driver->MajorFunction[IRP_MJ_CLOSE]=close;
	driver->MajorFunction[IRP_MJ_CLEANUP]=cleanup;
	driver->MajorFunction[IRP_MJ_READ]=read;
	driver->MajorFunction[IRP_MJ_WRITE]=write;

	return STATUS_SUCCESS;

error:
	if (dev_uname.Buffer) 
		RtlFreeUnicodeString(&dev_uname);
	if (link_uname.Buffer) 
		RtlFreeUnicodeString(&link_uname);


	DriverUnload(driver);
	return ret;
}
